#=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~=~
#                                      numgeo - GiD interface
#                       Copyright (C) 2023-today Jan Machacek, Rahul Dogra
#=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~=~
#
# History
# - 26.11.2023, J. Machacek - Initial version
#
# Memo:
# - get_domnode_attribute returns the current value of "n" (name) of $group
#
#
#=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~~=~=~

proc numgeo::WriteMesh { filename } {

    append filename "-mesh.inp"

    customlib::InitWriteFile $filename
	
	# xml document to get some tree data
	set root [$::gid_groups_conds::doc documentElement] ;
	
    customlib::WriteString "**=================================================================="
    customlib::WriteString "**                        numgeo input file"    
    customlib::WriteString "**=================================================================="
    customlib::WriteString "**"    
    customlib::WriteString "** Input file for the finite element code numgeo (Machacek & Staubach)"
    customlib::WriteString "**"   
    customlib::WriteString "** The input file was created by numgeo-GiD (Machacek & Dogra)"   
    customlib::WriteString "**"   
    customlib::WriteString "** numgeo is available from www.numgeo.de"   
    customlib::WriteString "** numgeo-GiD can be downloaded from github XXX" 
    customlib::WriteString "**"  
    customlib::WriteString "** Contact:"
    customlib::WriteString "** - Jan Machacek, jan.machacek@tu-darmstadt.de"
    customlib::WriteString "** - Rahul Dogra, rahul.dogra@tu-darmstadt.de"
    customlib::WriteString "**"
    customlib::WriteString "**=================================================================="
	
	
    #
	# get the dimensions of the model: 2D (plane-strain), 2D: Axissymmetric, 3D
	# 
	set dim_path {string(//container[@n="_dimensions_units_"]/value[@n="_dimensions_"]/@v)}
    set dim [$root selectNodes $dim_path]
	
    #
	# Nodes
	# 
	
    set units_mesh [gid_groups_conds::give_mesh_unit]
    customlib::WriteString ""
	customlib::WriteString "**"
	customlib::WriteString "** Nodes"
	customlib::WriteString "**"
	customlib::WriteString ""
    customlib::WriteString "*Node"
	if {$dim == "2D (plane-strain)"} {
      customlib::WriteCoordinates "%5d, %14.5e, %14.5e%.0s\n"
	} elseif {$dim == "2D: Axissymmetric"} {
      customlib::WriteCoordinates "%5d, %14.5e, %14.5e%.0s\n"
	} elseif {$dim == "3D"} {
      customlib::WriteCoordinates "%5d, %14.5e, %14.5e, %14.5e\n"
	}

	#
	# Elements
	#
	customlib::WriteString ""
	customlib::WriteString "**"
	customlib::WriteString "** Elements"
	customlib::WriteString "**"
	customlib::WriteString ""

	# assemble all triangular elements (linear and quadratic)
	set elements_t [GiD_Mesh list -element_type Triangle element]
	set elements_t3 {}
	set elements_t6 {}
	foreach i $elements_t {
	  set nodes [GiD_Mesh get element $i connectivities]
	  if {[llength $nodes] == 3} {
		lappend elements_t3 $i
	  } elseif {[llength $nodes] == 6} {
		lappend elements_t6 $i
	  }
	}

	# assemble all rectangular elements (linear and quadratic)
	set elements_q [GiD_Mesh list -element_type Quadrilateral element]
	set elements_q4 {}
	set elements_q8 {}
	set elements_q20 {}
	foreach i $elements_q {
	  set nodes [GiD_Mesh get element $i connectivities]
	  if {[llength $nodes] == 4} {
		lappend elements_q4 $i
	  } elseif {[llength $nodes] == 8} {
		lappend elements_q8 $i
	  } 
	}
	
	# assemble all rectangular elements (linear and quadratic)
	set elements_h [GiD_Mesh list -element_type Hexahedra element]
	set elements_h8 {}
	set elements_h20 {}
	foreach i $elements_h {
	  set nodes [GiD_Mesh get element $i connectivities]
	  if {[llength $nodes] == 8} {
		lappend elements_h8 $i
	  } elseif {[llength $nodes] == 20} {
		lappend elements_h20 $i
	  } 
	}

	set document [$::gid_groups_conds::doc documentElement]
	set xp [format_xpath {condition[@n="_element_types_"]/group}]
	# loop over all groups to which element types were assigned
	foreach group [$document selectNodes $xp] {
	  set group_name [get_domnode_attribute $group n]
	  set node [$group selectNodes "./value\[@n = '_element_formulation_'\]"]
	  set eltype [get_domnode_attribute $node v]
	  set node [$group selectNodes "./value\[@n = '_integration_'\]"]
	  set integration [get_domnode_attribute $node v]
	  set group_element_ids [GiD_EntitiesGroups get $group_name elements]

	  if {$dim == "2D (plane-strain)"} {
	    
        # solid element: 3 node triangle
		if {[llength $elements_t3] > 0} {
		  set keyword [return_eltype $dim $eltype $integration {Triangular} {3}]
		  # loop over all group members
		  for {set i 0} {$i < [llength $group_element_ids]} {incr i} {
			# get the element id and check if it is the correct element type
			set element_id [lindex $group_element_ids $i]
			if {[lsearch -exact $elements_t3 $element_id] >= 0} {
			  if {$i == 0} {
				# print header, only for first member of the group
				customlib::WriteString "*Element, type=$keyword"
			  }
			  # print connectivity of i-th element of group
			  lassign [GiD_Mesh get element $element_id connectivities] c1 c2 c3
			  customlib::WriteString "$element_id, $c1, $c2, $c3"
			}
		  }
		  customlib::WriteString ""
		}
		# solid element: 6 node triangle
		if {[llength $elements_t6] > 0} {
		  set keyword [return_eltype $dim $eltype $integration {Triangular} {6}]
		  for {set i 0} {$i < [llength $group_element_ids]} {incr i} {
			set element_id [lindex $group_element_ids $i]
			if {[lsearch -exact $elements_t6 $element_id] >= 0} {
			  if {$i == 0} {
				customlib::WriteString "*Element, type=$keyword"
			  }
			  lassign [GiD_Mesh get element $element_id connectivities] c1 c2 c3 c4 c5 c6
			  customlib::WriteString "$element_id, $c1, $c2, $c3, $c4, $c5, $c6"
			}
		  }
		  customlib::WriteString ""
		}
		# solid element: 4 node Quadrilateral
		if {[llength $elements_q4] > 0} {
		  set keyword [return_eltype $dim $eltype $integration {Quadrilateral} {4}]
		  for {set i 0} {$i < [llength $group_element_ids]} {incr i} {
			set element_id [lindex $group_element_ids $i]
			if {[lsearch -exact $elements_q4 $element_id] >= 0} {
			  if {$i == 0} {
				customlib::WriteString "*Element, type=$keyword"
			  }
			  lassign [GiD_Mesh get element $element_id connectivities] c1 c2 c3 c4
			  customlib::WriteString "$element_id, $c1, $c2, $c3, $c4"
			}
		  }
		  customlib::WriteString ""
		}
		# solid element: 8 node Quadrilateral
		if {[llength $elements_q8] > 0} {
		  set keyword [return_eltype $dim $eltype $integration {Quadrilateral} {8}]
		  for {set i 0} {$i < [llength $group_element_ids]} {incr i} {
			set element_id [lindex $group_element_ids $i]
			if {[lsearch -exact $elements_q8 $element_id] >= 0} {
			  if {$i == 0} {
				customlib::WriteString "*Element, type=$keyword"
			  }
			  lassign [GiD_Mesh get element $element_id connectivities] c1 c2 c3 c4 c5 c6 c7 c8
			  customlib::WriteString "$element_id, $c1, $c2, $c3, $c4, $c5, $c6, $c7, $c8"
			}
		  }
		  customlib::WriteString ""
		}
	  } elseif {$dim == "2D: Axissymmetric"} {
		# solid element: 3 node triangle
		if {[llength $elements_t3] > 0} {
		  set keyword [return_eltype $dim $eltype $integration {Triangular} {3}]
		  # loop over all group members
		  for {set i 0} {$i < [llength $group_element_ids]} {incr i} {
			# get the element id and check if it is the correct element type
			set element_id [lindex $group_element_ids $i]
			if {[lsearch -exact $elements_t3 $element_id] >= 0} {
			  if {$i == 0} {
				# print header, only for first member of the group
				customlib::WriteString "*Element, type=$keyword"
			  }
			  # print connectivity of i-th element of group
			  lassign [GiD_Mesh get element $element_id connectivities] c1 c2 c3
			  customlib::WriteString "$element_id, $c1, $c2, $c3"
			}
		  }
		  customlib::WriteString ""
		}
		# solid element: 6 node triangle
		if {[llength $elements_t6] > 0} {
		  set keyword [return_eltype $dim $eltype $integration {Triangular} {6}]
		  for {set i 0} {$i < [llength $group_element_ids]} {incr i} {
			set element_id [lindex $group_element_ids $i]
			if {[lsearch -exact $elements_t6 $element_id] >= 0} {
			  if {$i == 0} {
				customlib::WriteString "*Element, type=$keyword"
			  }
			  lassign [GiD_Mesh get element $element_id connectivities] c1 c2 c3 c4 c5 c6
			  customlib::WriteString "$element_id, $c1, $c2, $c3, $c4, $c5, $c6"
			}
		  }
		  customlib::WriteString ""
		}
		# solid element: 4 node Quadrilateral
		if {[llength $elements_q4] > 0} {
		  set keyword [return_eltype $dim $eltype $integration {Quadrilateral} {4}]
		  for {set i 0} {$i < [llength $group_element_ids]} {incr i} {
			set element_id [lindex $group_element_ids $i]
			if {[lsearch -exact $elements_q4 $element_id] >= 0} {
			  if {$i == 0} {
				customlib::WriteString "*Element, type=$keyword"
			  }
			  lassign [GiD_Mesh get element $element_id connectivities] c1 c2 c3 c4
			  customlib::WriteString "$element_id, $c1, $c2, $c3, $c4"
			}
		  }
		  customlib::WriteString ""
		}
		# solid element: 8 node Quadrilateral
		if {[llength $elements_q8] > 0} {
		  set keyword [return_eltype $dim $eltype $integration {Quadrilateral} {8}]
		  for {set i 0} {$i < [llength $group_element_ids]} {incr i} {
			set element_id [lindex $group_element_ids $i]
			if {[lsearch -exact $elements_q8 $element_id] >= 0} {
			  if {$i == 0} {
				customlib::WriteString "*Element, type=$keyword"
			  }
			  lassign [GiD_Mesh get element $element_id connectivities] c1 c2 c3 c4 c5 c6 c7 c8
			  customlib::WriteString "$element_id, $c1, $c2, $c3, $c4, $c5, $c6, $c7, $c8"
			}
		  }
		  customlib::WriteString ""
		}
	  } elseif {$dim == "3D"} {
		# solid element: 8 node Quadrilateral
		if {[llength $elements_h8] > 0} {
		  set keyword [return_eltype $dim $eltype $integration {Hexahedra} {8}]
		  for {set i 0} {$i < [llength $group_element_ids]} {incr i} {
			set element_id [lindex $group_element_ids $i]
			if {[lsearch -exact $elements_h8 $element_id] >= 0} {
			  if {$i == 0} {
				customlib::WriteString "*Element, type=$keyword"
			  }
			  lassign [GiD_Mesh get element $element_id connectivities] c1 c2 c3 c4 c5 c6 c7 c8
			  customlib::WriteString "$element_id, $c1, $c2, $c3, $c4, $c5, $c6, $c7, $c8"
			}
		  }
		  customlib::WriteString ""
		}
		# solid element: 20 node Quadrilateral
		if {[llength $elements_h20] > 0} {
		  set keyword [return_eltype $dim $eltype $integration {Hexahedra} {20}]
		  for {set i 0} {$i < [llength $group_element_ids]} {incr i} {
			set element_id [lindex $group_element_ids $i]
			if {[lsearch -exact $elements_h20 $element_id] >= 0} {
			  if {$i == 0} {
				customlib::WriteString "*Element, type=$keyword"
			  }
			  lassign [GiD_Mesh get element $element_id connectivities] c1 c2 c3 c4 c5 c6 c7 c8 c9 c10 c11 c12 c13 c14 c15 c16 c17 c18 c19 c20
			  customlib::WriteString "$element_id, $c1, $c2, $c3, $c4, $c5, $c6, $c7, $c8, $c9, $c10, $c11, $c12, $c13, $c14, $c15, $c16, $c17, $c18, $c19, $c20"
			}
		  }
		  customlib::WriteString ""
		}
	  }
	}

	#
	# Node sets
	#
	
	customlib::WriteString ""
	customlib::WriteString "**"
	customlib::WriteString "** Node sets"
	customlib::WriteString "**"
	customlib::WriteString ""
	
	set node_set_list {}
	
	set groups [GiD_Groups list]
	foreach group_name $groups {
		set group_entity_ids [GiD_EntitiesGroups get $group_name nodes]
		set entity_count [llength $group_entity_ids]
		if {$entity_count > 0} {
			lappend node_set_list $group_name
			set variables_per_line 16
			customlib::WriteString ""
			customlib::WriteString "*NSet, nset=$group_name"
			for {set i 0} {$i < $entity_count} {incr i $variables_per_line} {
				set end_index [expr {$i + $variables_per_line - 1}]
				set end_index [expr {$end_index >= $entity_count ? $entity_count - 1 : $end_index}]
				customlib::WriteString [join [lrange $group_entity_ids $i $end_index] ", "]
			}
		}
	}
	
	#
	# Element sets
	#

	customlib::WriteString ""
	customlib::WriteString "**"
	customlib::WriteString "** Element sets"
	customlib::WriteString "**"
	customlib::WriteString ""
	
	set groups [GiD_Groups list]
	foreach group_name $groups {
		set group_entity_ids [GiD_EntitiesGroups get $group_name elements]
		set entity_count [llength $group_entity_ids]
		if {$entity_count > 0} {
			set variables_per_line 16
			customlib::WriteString ""
			customlib::WriteString "*Elset, elset=$group_name"
			for {set i 0} {$i < $entity_count} {incr i $variables_per_line} {
				set end_index [expr {$i + $variables_per_line - 1}]
				set end_index [expr {$end_index >= $entity_count ? $entity_count - 1 : $end_index}]
				customlib::WriteString [join [lrange $group_entity_ids $i $end_index] ", "]
			}
		}
	}
	
	#
	# Surfaces
	#

	customlib::WriteString ""
	customlib::WriteString "**"
	customlib::WriteString "** Surfaces (generated from node sets)"
	customlib::WriteString "**"
	
	# we create surfaces from node sets
	# to keep the cost low we only create surfaces for node sets used to assign surface loads or contact conditions.
	
	set conditions [GiD_Groups list]
	foreach set_name $node_set_list {
      customlib::WriteString ""
	  customlib::WriteString "*Surface, name=$set_name, type = node"
	  customlib::WriteString "$set_name"
	}

	#
	# Close file
	#
	
    customlib::EndWriteFile
}